"""Models for Open Data Platform of Amsterdam."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any

import pytz

from .const import CORRECTIONS, FILTER_NAMES, FILTER_UNKNOWN


@dataclass
class ParkingSpot:
    """Object representing an ParkingSpot model response from the API."""

    spot_id: str
    spot_type: str | None
    spot_description: str | None

    street: str | None
    number: int | None
    orientation: str | None

    coordinates: list[float]

    @classmethod
    def from_json(cls: type[ParkingSpot], data: dict[str, Any]) -> ParkingSpot:
        """Return ParkingSpot object from a dictionary.

        Args:
        ----
            data: The JSON data from the API.

        Returns:
        -------
            An ParkingSpot object.
        """
        attr = data["properties"]
        regimes = attr["regimes"][0]
        return cls(
            spot_id=attr["id"],
            spot_type=attr["eType"] or None,
            spot_description=regimes["eTypeDescription"] or None,
            street=filter_unknown(attr["straatnaam"]),
            number=int(attr["aantal"]),
            orientation=filter_unknown(attr["type"]),
            coordinates=data["geometry"]["coordinates"][0],
        )


class VehicleType(str, Enum):
    """Enumeration representing the vehicle type."""

    BICYCLE = "bicycle"
    CAR = "car"
    TOURINGCAR = "touringcar"


class GarageCategory(str, Enum):
    """Enumeration representing the garage category."""

    GARAGE = "garage"
    PARK_AND_RIDE = "park_and_ride"


@dataclass
class Garage:
    """Object representing an Garage model response from the API."""

    garage_id: str
    garage_name: str
    vehicle: VehicleType
    category: GarageCategory
    state: str

    free_space_short: int
    free_space_long: int | None
    short_capacity: int
    long_capacity: int | None
    availability_pct: float | None

    longitude: float
    latitude: float
    updated_at: datetime

    @classmethod
    def from_json(cls: type[Garage], data: dict[str, Any]) -> Garage:
        """Return Garage object from a dictionary.

        Args:
        ----
            data: The JSON data from the API.

        Returns:
        -------
            An Garage object.
        """
        latitude, longitude = split_coordinates(str(data["geometry"]["coordinates"]))
        attr = data["properties"]
        return cls(
            garage_id=data["Id"],
            garage_name=correct_name(attr["Name"]),
            vehicle=get_vehicle_type(attr["Name"]),
            category=get_category(attr["Name"]),
            state=attr.get("State"),
            free_space_short=int(attr["FreeSpaceShort"]),
            free_space_long=set_long_parking(attr["FreeSpaceLong"]),
            short_capacity=int(attr["ShortCapacity"]),
            long_capacity=set_long_parking(attr["LongCapacity"]),
            availability_pct=calculate_pct(
                attr.get("FreeSpaceShort"),
                attr.get("ShortCapacity"),
            ),
            longitude=longitude,
            latitude=latitude,
            updated_at=datetime.strptime(
                attr["PubDate"],
                "%Y-%m-%dT%H:%M:%SZ",
            ).replace(tzinfo=pytz.timezone("UTC")),
        )


def set_long_parking(data: str) -> int | None:
    """Set the long parking capacity/free space value.

    Args:
    ----
        data: The data to be set.

    Returns:
    -------
        The long parking capacity/free space.
    """
    return None if not data else int(data)


def split_coordinates(data: str) -> tuple[float, float]:
    """Split the coordinate data in separate variables.

    Args:
    ----
        data: The data to be split.

    Returns:
    -------
        The coordinates.
    """
    longitude, latitude = data.split(", ")
    longitude = longitude.replace("[", "")
    latitude = latitude.replace("]", "")
    return float(latitude), float(longitude)


def calculate_pct(current: int, total: int) -> float | None:
    """Calculate the percentage of free parking spots.

    Args:
    ----
        current: The current amount of free parking spots.
        total: The total amount of parking spots.

    Returns:
    -------
        The percentage of free parking spots.
    """
    try:
        return round((float(current) / float(total)) * 100, 1)
    except ZeroDivisionError:
        return None


def get_category(name: str) -> GarageCategory:
    """Get the category from the garage name.

    Args:
    ----
        name: The name of the parking garage.

    Returns:
    -------
        The category name.
    """
    if "P+R" in name:
        return GarageCategory.PARK_AND_RIDE
    return GarageCategory.GARAGE


def get_vehicle_type(name: str) -> VehicleType:
    """Get the vehicle type from the garage name.

    Args:
    ----
        name: The name of the parking garage.

    Returns:
    -------
        The vehicle type.
    """
    if "-FP" in name:
        return VehicleType.BICYCLE
    if "PT" in name:
        return VehicleType.TOURINGCAR
    return VehicleType.CAR


def correct_name(name: str) -> str:
    """Change parking garage name for consistency if needed.

    Args:
    ----
        name: The name of the parking garage.

    Returns:
    -------
        The corrected name.
    """
    for value in FILTER_NAMES:
        # Remove parts from name string.
        name = name.replace(value, "")

    if "PR" in name:
        # Replace PR for P in name string.
        name = name.replace("PR", "P")

    if "FP-" in name:
        # Replace FP- for FP in name string.
        name = name.replace("FP-", "FP")

    if any(y in name for y in CORRECTIONS):
        # Add a 0 for consistency. (e.g. P3 -> P03)
        return name[:1] + "0" + name[1:]
    return name


def filter_unknown(data: str) -> str | None:
    """Filter unknown data from the API.

    Args:
    ----
        data: The data to be filtered.

    Returns:
    -------
        The filtered data.
    """
    if data in FILTER_UNKNOWN:
        return None
    return data



"""Asynchronous Python client providing Open Data information of Amsterdam."""
from __future__ import annotations

import asyncio
import json
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import TYPE_CHECKING, Any

import async_timeout
from aiohttp import ClientError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .const import FILTER_OUT
from .exceptions import (
    ODPAmsterdamConnectionError,
    ODPAmsterdamError,
    ODPAmsterdamResultsError,
)
from .models import Garage, ParkingSpot

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass
class ODPAmsterdam:
    """Main class for handling data fetchting from Open Data Platform of Amsterdam."""

    request_timeout: float = 15.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the Open Data Platform API of Amsterdam.

        Args:
        ----
            uri: Request URI, without '/', for example, 'status'
            method: HTTP method to use, for example, 'GET'
            params: Extra options to improve or limit the response.

        Returns:
        -------
            A Python dictionary (text) with the response from
            the Open Data Platform API of Amsterdam.

        Raises:
        ------
            ODPAmsterdamConnectionError: An error occurred while
                communicating with the Open Data Platform API of Amsterdam.
            ODPAmsterdamError: Received an unexpected response from
                the Open Data Platform API of Amsterdam.
        """
        version = metadata.version(__package__)
        url = URL.build(scheme="https", host="api.data.amsterdam.nl", path="/").join(
            URL(uri),
        )

        headers = {
            "Accept": "application/json, text/plain, application/geo+json",
            "User-Agent": f"PythonODPAmsterdam/{version}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with async_timeout.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    params=params,
                    headers=headers,
                    ssl=True,
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the Open Data Platform API."
            raise ODPAmsterdamConnectionError(msg) from exception
        except (ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with the Open Data Platform API."
            raise ODPAmsterdamConnectionError(msg) from exception

        types = ["application/json", "text/plain", "application/geo+json"]
        content_type = response.headers.get("Content-Type", "")
        if not any(item in content_type for item in types):
            text = await response.text()
            msg = "Unexpected content type response from the Open Data Platform API"
            raise ODPAmsterdamError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        return json.loads(await response.text())

    async def locations(
        self,
        limit: int = 10,
        parking_type: str = "",
    ) -> list[ParkingSpot]:
        """Get all the parking locations.

        Args:
        ----
            limit: The number of results to return.
            parking_type: The selected parking type number.

        Returns:
        -------
            A list of ParkingSpot objects.
        """
        locations = await self._request(
            "v1/parkeervakken/parkeervakken",
            params={"_pageSize": limit, "eType": parking_type, "_format": "geojson"},
        )
        return [ParkingSpot.from_json(item) for item in locations["features"]]

    async def all_garages(
        self,
        vehicle: str | None = None,
        category: str | None = None,
    ) -> list[Garage]:
        """Get all the garages.

        Returns
        -------
            A list of Garage objects.

        Raises
        ------
            ODPAmsterdamError: If the data is not valid.
        """
        data = await self._request("dcatd/datasets/9ORkef6T-aU29g/purls/1")

        try:
            results: list[Garage] = [
                Garage.from_json(item)
                for item in data["features"]
                if not any(x in item["properties"]["Name"] for x in FILTER_OUT)
            ]
        except KeyError as exception:
            msg = f"Got wrong data from the API: {exception}"
            raise ODPAmsterdamError(msg) from exception

        # Filter on vehicle type and category
        if vehicle:
            results = list(filter(lambda x: x.vehicle == vehicle, results))
        if category:
            results = list(filter(lambda x: x.category == category, results))
        return results

    async def garage(self, garage_id: str) -> Garage:
        """Get info from a single  garage.

        Args:
        ----
            garage_id: The ID of the garage.

        Returns:
        -------
            A garage object.

        Raises:
        ------
            ODPAmsterdamResultsError: When no results are found.
        """
        data = await self._request("dcatd/datasets/9ORkef6T-aU29g/purls/1")
        for item in data["features"]:
            if item["Id"] == garage_id:
                return Garage.from_json(item)
        msg = f"No garage was found with id - {garage_id}"
        raise ODPAmsterdamResultsError(msg)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The Open Data Platform Amsterdam object.
        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.
        """
        await self.close()



"""Asynchronous Python client providing Open Data information of Amsterdam."""
from __future__ import annotations

FILTER_OUT: list[str] = [
    "Dummy",
]

FILTER_NAMES: list[str] = [
    "CE-",
    "ZD-",
    "ZO-",
    "ZU-",
    "FJ212P34 ",
    "VRN-FJ212",
    "GRV020HNK ",
    " P ",
    " P4",
    " P5",
    " P21",
    " P22",
    " P23",
    " P24",
    "PR-",
    "DP-",
    "AM-",
]

FILTER_UNKNOWN: list[str] = [
    "ONBEKEND",
    "Onbekend",
]

CORRECTIONS: list[str] = [
    "P1 ",
    "P3 ",
]



"""Asynchronous Python client providing Open Data information of Amsterdam."""


class ODPAmsterdamError(Exception):
    """Generic Open Data Platform exception."""


class ODPAmsterdamConnectionError(ODPAmsterdamError):
    """Open Data Platform connection exception."""


class ODPAmsterdamResultsError(ODPAmsterdamError):
    """Open Data Platform Amsterdam results exception."""



"""Asynchronous Python client providing Open Data information of Amsterdam."""

from .exceptions import (
    ODPAmsterdamConnectionError,
    ODPAmsterdamError,
    ODPAmsterdamResultsError,
)
from .models import Garage, GarageCategory, ParkingSpot, VehicleType
from .odp_amsterdam import ODPAmsterdam

__all__ = [
    "Garage",
    "GarageCategory",
    "ODPAmsterdam",
    "ODPAmsterdamConnectionError",
    "ODPAmsterdamError",
    "ODPAmsterdamResultsError",
    "ParkingSpot",
    "VehicleType",
]
